/* src/c-interface-network.hh
 * 
 * Copyright 2012 by The University of Oxford
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License and
 * the GNU Lesser General Public License along with this program; if not,
 * write to the Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA, 02110-1301, USA.
 */

#include <string>

#ifdef USE_GUILE
std::string get_drug_via_wikipedia_and_drugbank_scm(const std::string &drugname);
#endif

#ifdef USE_PYTHON
std::string get_drug_via_wikipedia_and_drugbank_py(const std::string &drugname);
#endif

void curl_test_make_a_post();
void curl_post(const std::string &url, const std::string &post_string);
