grep -H == */*.cc | egrep -v ':  *if'| egrep -v ':  *while' | grep -v MAJOR | egrep -v ': *//' | egrep -v '{$' | egrep -v "  *\(" | grep -v '===' | grep -v std::cout | grep -v coot_wrap | grep -v ccp4mg | grep -v coot-utils/tree | egrep ';'
