/*
 * MoleculesToTriangles/CXXSurface/CXXTorusNode.h
 *
 * Copyright 2009 by Martin Noble, University of Oxford
 * Author: Martin Noble
 *
 * This file is part of Coot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA
 */

#ifndef CXXTorusNode_included
#define CXXTorusNode_included
#include "CXXCoord.h"
#include "mmdb2/mmdb_manager.h"

class CXXTorusNode{
private:
	mmdb::Atom* theAtom;
	double theta;
	double omega;
	CXXCoord<CXXCoord_ftype>crd;
	void init();
public:
		CXXTorusNode();
	CXXTorusNode(double the, double om);
	int setTheta(const double inTheta);
	int setOmega(const double inOmega);
	const CXXCoord<CXXCoord_ftype>&coord() const;
	const double getTheta() const;
	const double getOmega() const;
	int setCoord(const CXXCoord<CXXCoord_ftype>&);
	int setAtom(mmdb::Atom* anAtom);
	mmdb::Atom* getAtom() const;
};

#endif
