# a whole set of amercanisms
 
# an americanism
# deftexi set_rotation_center
set_rotation_center =  set_rotation_centre

# an americanism
# deftexi handle_read_draw_molecule_with_recenter
handle_read_draw_molecule_with_recenter = handle_read_draw_molecule_with_recentre

# an americanism
# deftexi set_rotation_center_size 
set_rotation_center_size = set_rotation_centre_size

# an americanism
# deftexi center_atom_label_status
center_atom_label_status = centre_atom_label_status

# an americanism
# deftexi set_last_map_color
set_last_map_color = set_last_map_colour

# an americanism
# deftexi center_of_mass
center_of_mass = centre_of_mass

# an americanism
# set deftexi set_swap_difference_map_colors
set_swap_difference_map_colors = set_swap_difference_map_colours

# an americanism
# deftexi set_font_color
set_font_color = set_font_colour

# an americanism
# deftexi set_dots_color
set_dots_color = set_dots_colour

# an americanism
invert_chiral_center = invert_chiral_centre


# some more from BL

# an americanism
# deftexi rotation_center
rotation_center = rotation_centre

# an americanism
# deftexi move_molecule_to_screen_center
move_molecule_to_screen_center = move_molecule_to_screen_centre
