

/* functions used in callbacks.c */

#ifndef BEGIN_C_DECLS
#ifdef __cplusplus
#define BEGIN_C_DECLS extern "C" {
#define END_C_DECLS }
#else
#define BEGIN_C_DECLS
#define END_C_DECLS
#endif
#endif

BEGIN_C_DECLS
int crankshaft_peptide_rotation_optimization_intermediate_atoms();
void curlew();

END_C_DECLS
