/*
* @author Filomeno Sanchez
* 
* Copyright 2023, University of York
* All rights reserved.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as published
* by the Free Software Foundation; either version 3 of the License, or (at
* your option) any later version.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
* 02110-1301, USA
*/

#include "molecules-container.hh"
#include "test-molecules-container.hh"

int test_something_filo(molecules_container_t &mc);

int test_get_diff_map_peaks(molecules_container_t &mc);

int test_non_drawn_bond_multi_cid_2(molecules_container_t &mc);

int test_change_chain_id_1(molecules_container_t &molecules_container);

int test_change_rotamer(molecules_container_t &molecules_container);

int test_import_ligands_with_same_name_and_animated_refinement(molecules_container_t &molecules_container);
